/*****************************************************************************
 * Copyright (C) The Apache Software Foundation. All rights reserved.        *
 * ------------------------------------------------------------------------- *
 * This software is published under the terms of the Apache Software License *
 * version 1.1, a copy of which has been included with this distribution in  *
 * the LICENSE file.                                                         *
 *****************************************************************************/

package org.apache.batik.transcoder;

import java.util.Map;
import java.util.Set;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.batik.refimpl.transcoder.BatikHints;

/**
 * The <tt>TranscodingHints</tt> class contains transcoding hints that
 * can be used by <tt>Transcoder</tt> class.
 *
 * @author <a href="mailto:Thierry.Kormann@sophia.inria.fr">Thierry Kormann</a>
 * @version $Id: TranscodingHints.java,v 1.2 2000/11/21 15:01:00 tkormann Exp $
 */
public class TranscodingHints implements Map, Cloneable {

    /**
     * Background color key.
     */
    public static final Key KEY_BACKGROUND =
        BatikHints.KEY_BACKGROUND;

    /**
     * The XML parser classname key.
     */
    public static final Key KEY_XML_PARSER_CLASSNAME =
        BatikHints.KEY_XML_PARSER_CLASSNAME;

    /**
     * The default viewport.
     */
    public static final Key KEY_DEFAULT_VIEWPORT =
        BatikHints.KEY_DEFAULT_VIEWPORT;

    /**
     * The <tt>GVTBuilder</tt> implementation to use.
     */
    public static final Key KEY_GVT_BUILDER =
        BatikHints.KEY_GVT_BUILDER;

    /** The transcoding hints. */
    HashMap hintMap = new HashMap(7);

    /**
     * Constructs a new empty object.
     */
    public TranscodingHints() {
    }

    /**
     * Constructs a new object with keys and values initialized from
     * the specified Map object (which may be null).
     *
     * @param init a map of key/value pairs to initialize the hints
     *          or null if the object should be empty
     */
    public TranscodingHints(Map init) {
        if (init != null) {
            hintMap.putAll(init);
        }
    }

    /**
     * Constructs a new object with the specified key/value pair.
     *
     * @param key the key of the particular hint property
     * @param value the value of the hint property specified with
     * <tt>key</tt>
     */
    public TranscodingHints(Key key, Object value) {
        hintMap.put(key, value);
    }

    /**
     * Returns the number of key-value mappings in this
     * <tt>TranscodingHints</tt>.
     */
    public int size() {
        return hintMap.size();
    }

    /**
     * Returns <tt>true</tt> if this <tt>TranscodingHints</tt> contains
     * no key-value mappings, false otherwise.
     */
    public boolean isEmpty() {
        return hintMap.isEmpty();
    }

    /**
     * Returns <tt>true</tt> if this <tt>TranscodingHints</tt> contains a
     * mapping for the specified key, false otherwise.
     *
     * @param key key whose present in this <tt>TranscodingHints</tt>
     * is to be tested.
     */
    public boolean containsKey(Object key) {
        return hintMap.containsKey(key);
    }

    /**
     * Returns <tt>true</tt> if this <tt>TranscodingHints</tt> maps
     * one or more keys to the specified value.
     *
     * @param value value whose presence in this <tt>TranscodingHints</tt>
     * is to be tested.
     */
    public boolean containsValue(Object value) {
        return hintMap.containsValue(value);
    }

    /**
     * Returns the value to which the specified key is mapped.
     *
     * @param key a trancoding hint key
     * @exception ClassCastException key is not of type
     * <tt>TranscodingHints.Key</tt>
     */
    public Object get(Object key) {
        return hintMap.get((Key) key);
    }

    /**
     * Maps the specified <tt>key</tt> to the specified <tt>value</tt>
     * in this <tt>TranscodingHints</tt> object.
     *
     * @param key the trancoding hint key.
     * @param value the trancoding hint value.
     * @exception <tt>IllegalArgumentException</tt> value is not
     * appropriate for the specified key.
     * @exception ClassCastException key is not of type
     * <tt>TranscodingHints.Key</tt>
     */
    public Object put(Object key, Object value) {
        if (!((Key) key).isCompatibleValue(value)) {
            throw new IllegalArgumentException(value+
                                               " incompatible with"+
                                               key);
        }
        return hintMap.put((Key) key, value);
    }

    /**
     * Adds all of the keys and corresponding values from the
     * specified <tt>TranscodingHints</tt> object to this
     * <tt>TranscodingHints</tt> object.
     */
    public void add(TranscodingHints hints) {
        hintMap.putAll(hints);
    }

    /**
     * Clears this <tt>TranscodingHints</tt> object of all key/value pairs.
     */
    public void clear() {
        hintMap.clear();
    }

    /**
     * Removes the key and its corresponding value from this
     * <tt>TranscodingHints</tt> object.
     *
     * @param key the trancoding hints key that needs to be removed
     * @exception ClassCastException key is not of type
     * <tt>TranscodingHints.Key</tt>
     */
    public Object remove(Object key) {
        return hintMap.remove((Key) key);
    }

    /**
     * Copies all of the mappings from the specified <tt>Map</tt>
     * to this <tt>TranscodingHints</tt>.
     *
     * @param t mappings to be stored in this <tt>TranscodingHints</tt>.
     * @exception ClassCastException key is not of type
     * <tt>TranscodingHints.Key</tt>
     */
    public void putAll(Map m) {
        if (m instanceof TranscodingHints) {
            hintMap.putAll(((TranscodingHints) m).hintMap);
        } else {
            Iterator iter = m.entrySet().iterator();
            while (iter.hasNext()) {
                Map.Entry entry = (Map.Entry) iter.next();
                put(entry.getKey(), entry.getValue());
            }
        }
    }

    /**
     * Returns a <tt>Set</tt> view of the Keys contained in this
     * <tt>TranscodingHints</tt>.
     */
    public Set keySet() {
        return hintMap.keySet();
    }

    /**
     * Returns a <tt>Collection</tt> view of the values contained in this
     * <tt>TranscodingHints</tt>.
     */
    public Collection values() {
        return hintMap.values();
    }

    /**
     * Returns a <tt>Set</tt> view of the mappings contained
     * in this <tt>TranscodingHints</tt>.
     */
    public Set entrySet() {
        return Collections.unmodifiableMap(hintMap).entrySet();
    }

    /**
     * Compares the specified <tt>Object</tt> with this
     * <tt>TranscodingHints</tt> for equality.
     */
    public boolean equals(Object o) {
        if (o instanceof TranscodingHints) {
            return hintMap.equals(((TranscodingHints) o).hintMap);
        } else if (o instanceof Map) {
            return hintMap.equals(o);
        }
        return false;
    }

    /**
     * Returns the hash code value for this <tt>TranscodingHints</tt>.
     */
    public int hashCode() {
        return hintMap.hashCode();
    }

    /**
     * Creates a clone of this <tt>TranscodingHints</tt> object that
     * has the same contents as this <tt>TranscodingHints</tt> object.
     */
    public Object clone() {
        TranscodingHints rh;
        try {
            rh = (TranscodingHints) super.clone();
            if (hintMap != null) {
                rh.hintMap = (HashMap) hintMap.clone();
            }
        } catch (CloneNotSupportedException e) {
            // should not be reached
            throw new InternalError();
        }
        return rh;
    }


    /**
     * Defines the base type of all keys used to control various
     * aspects of the trancoding operations.
     */
    public abstract static class Key {

        /**
         * Returns true if the specified object is a valid value for this key.
         */
        public abstract boolean isCompatibleValue(Object val);
    }
}
